#!/usr/bin/python

# CRIMSON
# Copyright (c) 2006, Stephen Fisher, Susan Davidson, and Junhyong Kim, 
# University of Pennsylvania.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA.
#
# startup.py - initialize console
#
# @author  Stephen Fisher
# @version $Id: nex2walrus.py,v 1.13 2007/05/22 21:17:10 fisher Exp $

# This assumes that the entire tree is on one line.

# todo:
#   deal with RGB conversion

# load standard definitions
from header import *

import string
from org.python.modules import re
from math import *

def export(tree, filename, *args):
    """export(tree, filename [, include_ID, include_StemLen, include_tempDepth, include_coloring]):
    This will create a Walrus input file from the input tree.  The
    additional boolean arguments, flag whether to include the
    respective attributes in the output file.
    """
    file = open(filename, "w")
    if file == None: return

    numSpecies = tree.getNumSpecies()
    minStemLen = tree.getMinStemLength()
    maxStemLen = tree.getMaxStemLength()
    minTempDepth = tree.getMinTempDepth()
    maxTempDepth = tree.getMaxTempDepth()
    
    printMsg("Count: " + str(numSpecies))
    printMsg("Min Stem Length: " + str(minStemLen) + " Max Stem Length: " + str(maxStemLen))
    printMsg("Min TempDepth: " + str(minTempDepth) + " Max TempDepth: " + str(maxTempDepth))

    if minStemLen == 0: minStemLen = 0.01
    logMinStemLen = -log(minStemLen)
    logMaxStemLen = -log(maxStemLen)
    if minTempDepth == 0: minTempDepth = 0.01
    logMinTempDepth = -log(minTempDepth)
    logMaxTempDepth = -log(maxTempDepth)
    
    if len(args) == 0:
        incID = 0
        incStemLen = 0
        incTempDepth = 0
        incBLColor = 0
        incRLColor = 0
    else:
        incID = args[0]
        incStemLen = args[1]
        incTempDepth = args[2]
        if args[3] == 1:
            incBLColor = 1
            incRLColor = 1
        else:
            incBLColor = 0
            incRLColor = 0
        
    file.write("Graph\n")
    file.write("{\n")

    printMsg("Adding header info...")
    file.write("### metadata ###\n")
    file.write("@name=\"ROOT\";\n")
    file.write("@description=\"Export from Crimson\";\n")
    stmp = "@numNodes=" + str(numSpecies) + ";\n" 
    file.write(stmp)
    stmp = "@numLinks=" + str(numSpecies-1) + ";\n"
    file.write(stmp)
    file.write("@numPaths=0;\n")
    file.write("@numPathLinks=0;\n")
    file.write("\n")
    
    printMsg("Adding links...")
    file.write("### structural data ###\n")
    file.write("@links=[\n")
    _addLinks(tree.getRoot(), file, 0)
    file.write("];\n")
    file.write("@paths=;\n")

    printMsg("Adding attributes...")
    file.write("### attribute data ###\n")
    file.write("@enumerations=;\n")
    file.write("@attributeDefinitions=[\n")
    file.write("{\n")
    file.write("      @name=$root;\n")
    file.write("      @type=bool;\n")
    file.write("      @default=|| false ||;\n")
    file.write("      @nodeValues=[ { @id=0; @value=T; } ];\n")
    file.write("      @linkValues=;\n")
    file.write("      @pathValues=;\n")
    file.write("}\n")
    file.write(", {\n")
    file.write("      @name=$tree_link;\n")
    file.write("      @type=bool;\n")
    file.write("      @default=|| true ||;\n")
    file.write("      @nodeValues=;\n")
    file.write("      @linkValues=;\n")
    file.write("      @pathValues=;\n")
    file.write("}\n")

    if incID:
        printMsg("Adding species IDs...")
        file.write(", {\n")
        file.write("      @name=$Label;\n")
        file.write("      @type=string;\n")
        file.write("      @default=;\n")
        file.write("      @nodeValues=[\n")
        _addIDs(tree.getRoot(), file, 0)
        file.write("];\n")
        file.write("      @linkValues=;\n")
        file.write("      @pathValues=;\n")
        file.write("}\n")

    if incStemLen:
        printMsg("Adding species stem lengths...")
        file.write(", {\n")
        file.write("      @name=$Stem_Length;\n")
        file.write("      @type=double;\n")
        file.write("      @default=;\n")
        file.write("      @nodeValues=[\n")
        _addStemLens(tree.getRoot(), file, 0)
        file.write("];\n")
        file.write("      @linkValues=;\n")
        file.write("      @pathValues=;\n")
        file.write("}\n")
        
    if incTempDepth:
        printMsg("Adding species temporal depth...")
        file.write(", {\n")
        file.write("      @name=$TempDepth;\n")
        file.write("      @type=double;\n")
        file.write("      @default=;\n")
        file.write("      @nodeValues=[\n")
        _addTempDepth(tree.getRoot(), file, 0)
        file.write("];\n")
        file.write("      @linkValues=;\n")
        file.write("      @pathValues=;\n")
        file.write("}\n")
            
    if incBLColor:
        printMsg("Adding stem length coloring...")
        file.write(", {\n")
        file.write("      @name=$color_Stem_Length;\n")
        file.write("      @type=int;\n")
        file.write("      @default=;\n")
        file.write("      @nodeValues=[\n")
        _addNodeStemLenColors(tree.getRoot(), file, logMinStemLen, (logMaxStemLen-logMinStemLen), 0)
        file.write("];\n")
        file.write("      @linkValues=[\n")
        _addLinkStemLenColors(tree.getRoot(), file, logMinStemLen, (logMaxStemLen-logMinStemLen), -1)
        file.write("];\n")
        file.write("      @pathValues=;\n")
        file.write("}\n")
                
    if incRLColor:
        printMsg("Adding temporal depth coloring...")
        file.write(", {\n")
        file.write("      @name=$Color_TempDepth;\n")
        file.write("      @type=int;\n")
        file.write("      @default=;\n")
        file.write("      @nodeValues=[\n")
        _addNodeTempDepthColors(tree.getRoot(), file, logMinTempDepth, (logMaxTempDepth-logMinTempDepth), 0)
        file.write("];\n")
        file.write("      @linkValues=[\n")
        _addLinkTempDepthColors(tree.getRoot(), file, logMinTempDepth, (logMaxTempDepth-logMinTempDepth), -1)
        file.write("];\n")
        file.write("      @pathValues=;\n")
        file.write("}\n")
                
    file.write("];\n")

    # printMsg("Adding qualifiers...")
    file.write("@qualifiers=[\n")
    file.write("{\n")
    file.write("      @type=$spanning_tree;\n")
    file.write("      @name=$sample_spanning_tree;\n")
    file.write("      @description=;\n")
    file.write("      @attributes=[\n")
    file.write("         { @attribute=0; @alias=$root; },\n")
    file.write("         { @attribute=1; @alias=$tree_link; }\n")
    file.write("      ];\n")
    file.write("}\n")
    file.write("];\n")

    # printMsg("Adding visualization config...")
    file.write("### visualization hints ###\n")
    file.write("@filters=;\n")
    file.write("@selectors=;\n")
    file.write("@displays=;\n")
    file.write("@presentations=;\n")

    # printMsg("Adding interface config...")
    file.write("### interface hints ###\n")
    file.write("@presentationMenus=;\n")
    file.write("@displayMenus=;\n")
    file.write("@selectorMenus=;\n")
    file.write("@filterMenus=;\n")
    file.write("@attributeMenus=;\n")
    
    file.write("}\n")
    file.flush()

# recursively adds links
def _addLinks(species, file, pLabel):
    # if is leaf, then nothing to print
    if species.isLeaf(): return pLabel
    
    cLabel = pLabel + 1
    for child in species.getChildren():
        sTmp = ""
        if not species.isRoot() or cLabel > 1: sTmp = ",\n"
        sTmp += "{ @source=" + str(pLabel) + "; @destination=" + str(cLabel) + "; }"
        file.write(sTmp)
        
        # print out this child's children
        lTmp = _addLinks(child, file, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp
        
    return cLabel

# recursively adds species IDs
def _addIDs(species, file, pLabel):
    sTmp = ""
    if not species.isRoot(): sTmp = ",\n"
    sTmp += "{ @id=" + str(pLabel) + "; @value=\"" + str(species.getID()) + "\"; }"
    file.write(sTmp)

    # if is leaf, then done
    if species.isLeaf(): return pLabel

    cLabel = pLabel + 1
    for child in species.getChildren():
        # print out this child's children
        lTmp = _addIDs(child, file, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp
        
    return cLabel

# recursively adds species stem lengths
def _addStemLens(species, file, pLabel):
    sTmp = ""
    if not species.isRoot(): sTmp = ",\n"
    sTmp += "{ @id=" + str(pLabel) + "; @value=" + str(species.getStemLength()) + "; }"
    file.write(sTmp)

    # if is leaf, then done
    if species.isLeaf(): return pLabel

    cLabel = pLabel + 1
    for child in species.getChildren():
        # print out this child's children
        lTmp = _addStemLens(child, file, cLabel)

        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel

# recursively adds species temporal depth
def _addTempDepth(species, file, pLabel):
    sTmp = ""
    if not species.isRoot(): sTmp = ",\n"
    sTmp += "{ @id=" + str(pLabel) + "; @value=" + str(species.getTempDepth()) + "; }"
    file.write(sTmp)

    # if is leaf, then done
    if species.isLeaf(): return pLabel

    cLabel = pLabel + 1
    for child in species.getChildren():
        # print out this child's children
        lTmp = _addTempDepth(child, file, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel


# recursively adds species stem lengths as node colors
def _addNodeStemLenColors(species, file, min, diff, pLabel):
    if species.getStemLength() == 0: value = 255
    else: value = int((((-log(species.getStemLength())) - min) / diff) * 255)
    # value = -java.awt.Color(value, 255-value, 255).getRGB()
    value = _convertRGB(value, 255-value, 255)
    
    sTmp = ""
    if not species.isRoot(): sTmp = ",\n"
    sTmp += "{ @id=" + str(pLabel) + "; @value=" + str(value) + "; }"
    file.write(sTmp)
    
    # if is leaf, then done
    if species.isLeaf(): return pLabel
    
    cLabel = pLabel + 1
    for child in species.getChildren():
        # print out this child's children
        lTmp = _addNodeStemLenColors(child, file, min, diff, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel

# recursively adds species stem lengths as link colors
def _addLinkStemLenColors(species, file, min, diff, pLabel):
    # if is leaf, then nothing to print
    if species.isLeaf(): return pLabel
    
    cLabel = pLabel + 1
    for child in species.getChildren():
        sTmp = ""
        if not species.isRoot() or cLabel >= 1: sTmp = ",\n"
        
        if child.getStemLength() == 0: value = 255
        else: value = int((((-log(child.getStemLength())) - min) / diff) * 255)
        # value = -java.awt.Color(value, 255-value, 255).getRGB()
        value = _convertRGB(value, 255-value, 255)
        
        sTmp += "{ @id="+ str(cLabel) + "; @value=" + str(value) + "; }"
        file.write(sTmp)
        
        # print out this child's children
        lTmp = _addLinkStemLenColors(child, file, min, diff, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel

# recursively adds species temporal depth as node colors
def _addNodeTempDepthColors(species, file, min, diff, pLabel):
    if species.getTempDepth() == 0: value = 255
    else: value = int((((-log(species.getTempDepth())) - min) / diff) * 255)
    # value = -java.awt.Color(value, 255-value, 255).getRGB()
    value = _convertRGB(value, 255-value, 255)
    
    sTmp = ""
    if not species.isRoot(): sTmp = ",\n"
    sTmp += "{ @id=" + str(pLabel) + "; @value=" + str(value) + "; }"
    file.write(sTmp)
    
    # if is leaf, then done
    if species.isLeaf(): return pLabel
    
    cLabel = pLabel + 1
    for child in species.getChildren():
        # print out this child's children
        lTmp = _addNodeTempDepthColors(child, file, min, diff, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel

# recursively adds species temporal depth as link colors
def _addLinkTempDepthColors(species, file, min, diff, pLabel):
    # if is leaf, then nothing to print
    if species.isLeaf(): return pLabel
    
    cLabel = pLabel + 1
    for child in species.getChildren():
        sTmp = ""
        if not species.isRoot() or cLabel >= 1: sTmp = ",\n"
        
        if child.getTempDepth() == 0: value = 255
        else: value = int((((-log(child.getTempDepth())) - min) / diff) * 255)
        # value = -java.awt.Color(value, 255-value, 255).getRGB()
        value = _convertRGB(value, 255-value, 255)
        
        sTmp += "{ @id="+ str(cLabel) + "; @value=" + str(value) + "; }"
        file.write(sTmp)
        
        # print out this child's children
        lTmp = _addLinkTempDepthColors(child, file, min, diff, cLabel)
        
        # if this child is leaf then increment cLabel, else use
        # the value returned
        if lTmp == cLabel: cLabel += 1
        else: cLabel = lTmp

    return cLabel

def _convertRGB(r, g, b):
    return (r << 24) | (g << 16) | (b << 8)
